/*
 * erl_marshal.c
 */

#include <stdlib.h>
#include <string.h>
#include "erl_marshal.h"

#ifdef HAVE_MACHINE_ENDIAN_H
#include <machine/endian.h>
#endif

#ifdef HAVE_ENDIAN_H
#include <endian.h>
#endif

static void add_undecorated (erl_message *message, const void *data,
			     size_t len, request_rec *r);

static const void * read_undecorated (erl_message *lessage, size_t len);

void
init_message (erl_message *message)
{
  message->index = 0;
  message->receiving = NULL;
  message->length = 0;
}

static void
add_undecorated (erl_message *message, const void *data,
		 size_t len, request_rec *r)
{
  if (message->index >= MAXIOVECS) {
    /* FIXME */
    abort ();
  }
  message->iovecs[message->index].iov_base = ap_palloc (r->pool, len);
  memcpy (message->iovecs[message->index].iov_base, data, len);
  message->iovecs[message->index].iov_len = len;
  message->length += len;
  message->index++;
}

void
add_raw (erl_message *message, const void *data, size_t len, request_rec *r)
{
  add_integer (message, len, r);
  add_undecorated (message, data, len, r);
}

void
add_integer (erl_message *message, int value, request_rec *r)
{
#if defined (BYTE_ORDER) && BYTE_ORDER == BIG_ENDIAN
  /* We can optimize it if we know for sure that we are big endian */
  add_undecorated (message, &value, 4, r);
#else
  /* Reorder the bytes if needed */
  unsigned char big_endian_value[4];
  big_endian_value[0] = value >> 24;
  big_endian_value[1] = (value >> 16) & 0xff;
  big_endian_value[2] = (value >> 8) & 0xff;
  big_endian_value[3] = value & 0xff;
  add_undecorated (message, big_endian_value, 4, r);
#endif
}

static const void *
read_undecorated (erl_message *message, size_t len)
{
  void *result = &message->receiving[message->index];
  message->index += len;
  if (message->index > message->length) {
    init_message (message);
  }
  return result;
}

const void *
read_raw (erl_message *message, size_t *len)
{
  *len = read_integer (message);
  return read_undecorated (message, *len);
}

const char *
read_string (erl_message *message, request_rec *r)
{
  size_t len;
  const char *str = read_raw (message, &len);
  char *result = ap_pcalloc (r->pool, len + 1);
  memcpy (result, str, len);
  return result;
}

int
read_integer (erl_message *message)
{
#if defined (BYTE_ORDER) && BYTE_ORDER == BIG_ENDIAN
  /* Same note as above concerning optimization */
  return *((int *) read_undecorated (message, 4));
#else
  unsigned const char *big_endian_value = read_undecorated (message, 4);
  return (big_endian_value[0] << 24) |
    (big_endian_value[1] << 16) |
    (big_endian_value[2] << 8) |
    big_endian_value[3];
#endif
}

int
send_message (int fd, erl_message *message, int add_length, request_rec *r)
{
  int state = 0;
  int i;

  if (add_length) {
    /* Send a first message containing the length */
    erl_message length_message;
    init_message (&length_message);
    add_integer (&length_message, message->length, r);
    send_message (fd, &length_message, 0, r);
  }
  if (writev (fd, message->iovecs, message->index) < message->length) {
    state = 1;
  }
  init_message (message);
  return state;
}

int
read_message (int fd, erl_message *message, size_t receive_length,
	      request_rec *r)
{
  int n;
  unsigned char *index;
  size_t count = 0;

  if (!receive_length) {
    read_message (fd, message, 4, r);
    receive_length = read_integer (message);
  }

  message->length = receive_length;
  index = message->receiving = ap_palloc (r->pool, receive_length);
  while (count < receive_length) {
    n = read (fd, index, (receive_length - count));
    if (n <= 0) {
      init_message (message);
      return 1;
    }
    index += n;
    count += n;
  }
  return 0;
}
