/*
 * erl_scan.c
 */

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "erl_scan.h"

static char *look_for_lt (const char *text);
/* Look for < in text and return its position */

static char *look_for_qm (const char *text);
/* Look for ? in text and return its position */

static const char *look_for_gt (const char *text);
/* Look for > in text (we are in a tag, so we have to properly escape
 * strings) and return its position. */

static const char *look_for_eos (const char *text, char end_char);
/* Look for the end of string ending with char end_char */

static int count_spaces (const char *text);
/* Return the number of white character in text */

static const char *get_char (const char *text, char ch);
/* Return the address of char, which must be the next non-blank character,
 * NULL is returned otherwise. */

static const char *get_identifier (const char *text, int *spaces);
/* Get the next identifier, that is a sequence terminated by a non
 * alphanumeric character. It contains at least one character, even if
 * it is not an alphanumeric one, but not a space (they are counted
 * in spaces). */

static char *
look_for_lt (const char *text)
{
  return strchr (text, '<');
}

static char *
look_for_qm (const char *text)
{
  return index (text, '?');
}

static const char *
look_for_gt (const char *text)
{
  for (; *text; text++) {
    if (*text == '>') {
      return text;
    } else if (*text == '"') {
      text = look_for_eos (text + 1, '"');
    } else if (*text == '\'') {
      text = look_for_eos (text + 1, '\'');
    }
  }
  return NULL;
}

static const char *
look_for_eos (const char *text, char end_char)
{
  for (; *text; text++) {
    if (*text == end_char) {
      return text;
    } else if (*text == '\\') {
      text++;
    }
  }
  return text - 1;
}

static int
count_spaces (const char *text)
{
  int count;
  for (count = 0; isspace (*text); text++, count++);
  return count;
}

static const char *get_char (const char *text, char ch)
{
  int spaces = count_spaces (text);
  if (text[spaces] == ch) {
    return &text[spaces];
  } else {
    return NULL;
  }
}

static const char *
get_identifier (const char *text, int *spaces)
{
  const char *start, *index;
  static char identifier [80];
  int length;
  *spaces = count_spaces (text);
  start = text + *spaces;
  for (length = 1, index = start + 1;
       length + 1 < sizeof identifier && isalnum (*index);
       index++, length ++);
  memcpy (identifier, start, length);
  identifier[length] = '\0';
  return identifier;
}

int
lookup_erl (const char *text, char **begin, char **end,
	    char **content_begin, char **content_end,
	    erl_option **options)
{
  const char *current = text;
  const char *id;
  int short_form;
  int spaces;

  /* FIXME: options handling */
  *options = NULL;

  for (;;) {
    *begin = look_for_lt (current);
    if (*begin == NULL) {
      return 0;
    }
    id = get_identifier (*begin + 1, &spaces);
    if (!strcmp (id, "?")) {
      short_form = 1;
      *content_begin = *begin + spaces + 2;
      break;
    } else if (!strcmp (id, "erl")) {
      short_form = 0;
      *content_begin = (char *) look_for_gt (*begin + spaces + 3) + 1;
      break;
    } else {
      current = look_for_gt (*begin + spaces + strlen (id));
      if (current == NULL) {
	return 0;
      }
      current++;
    }
  }

  current = *content_begin;
  if (short_form) {
    for (;;) {
      current = look_for_qm (current);
      if (current == NULL) {
	return 0;
      }
      *end = (char *) get_char (current + 1, '>');
      if (*end) {
	*content_end = (char *) current - 1;
	return 1;
      } else {
	current++;
      }
    }
  } else {
    for (;;) {
      current = look_for_lt (current);
      if (current == NULL) {
	return 0;
      }
      *content_end = (char *) current - 1;
      id = get_char (current + 1, '/');
      if (id == NULL) {
	current++;
	continue;
      }
      current = id;
      id = get_identifier (id + 1, &spaces);
      if (!strcmp (id, "erl")) {
	*end = (char *) get_char (current + spaces + 4, '>');
	if (*end != NULL) {
	  return 1;
	}
      }
    }
  }
}
