-module (eerl_handler).
-export ([start/0, start_link/0, loop/2, find_variable/1, find_input/1,
	  fetch_variable/1, fetch_input/1, set_content_type/1,
	  set_return_code/1, set_return_code_immediate/1,
	  get_cookies/0, find_cookie/1, fetch_cookie/1, add_header/2,
	  redirect/1, set_only_headers/0, self_url/0, handler/2,
          variables/0, inputs/0, set_debug/1]).
-compile (export_all).

-define (PORT, 4163).

start () ->
    {ok, Sock} = gen_tcp:listen (?PORT, [binary,
					 {packet, 4},
					 {active, false},
					 {reuseaddr, true}]),
    PID = spawn (?MODULE, loop, [Sock, false]),
    register (eerl_handler_acceptor, PID),
    {ok, PID}.

start_link () ->
    {ok, PID} = start (),
    link (PID),
    {ok, PID}.

loop (Sock, Debug) ->
    {ok, NS} = gen_tcp:accept (Sock),
    NewDebug = receive
		   {set_debug, D} -> D
	       after 0 -> Debug
	       end,
    spawn (?MODULE, handler, [NS, NewDebug]),
    ?MODULE:loop (Sock, NewDebug).

encapsulate (Data) when binary (Data) ->
    Length = size (Data),
    <<Length:32, Data/binary>>;

encapsulate (Data) when list (Data) ->
    encapsulate (list_to_binary (Data)).

handler (Sock, Debug) ->
    case gen_tcp:recv (Sock, 0) of
	{ok, Request} ->
	    if Debug == true ->
		    io:format ("Received request: {~p, ~p, ~p}~n",
			       tuple_to_list (analyze_request (Request)));
	       true  ->
		    ok
	    end,
	    Result = handle_request (Request),
	    if Debug == true ->
		    io:format ("Answer is: ~p~n", [Result]);
	       true  ->
		    ok
	    end,
	    RC = case get (return_code) of
		     undefined -> 0;
		     X         -> X
		 end,
	    ContentType = encapsulate (case get (contenttype) of
					   undefined -> "";
					   Y         -> lower (Y)
				       end),
	    Data = encapsulate (Result),
	    Headers = make_headers (get (headers)),
	    gen_tcp:send (Sock, encapsulate (<<RC:32, ContentType/binary,
					     Headers/binary, Data/binary>>)),
	    ?MODULE:handler (Sock, Debug);
	_ ->
	    gen_tcp:close (Sock)
    end.

analyze_request (Buffer) ->
    {Input, R1} = read_string (Buffer),
    {Env, R2} = read_strings (R1),
    {Code, R3} = read_string (R2),
    {Input, Env, Code}.

handle_request (Buffer) ->
    {Input, Env, Code} = analyze_request (Buffer),
    put (input, Input),
    put (env, Env),
    put (return_code, undefined),
    put (contenttype, undefined),
    put (headers, []),
    RawResult = case catch eval (Code ++ ". ") of
		    {'EXIT', Error} ->
			io_lib:format
			  ("<font color=#ff0000><b>Exception received: " ++
			   "~p</b></font>~n", [Error]);
		    R -> R
		end,
    Result = if
		 integer (RawResult) -> integer_to_list (RawResult);
		 atom (RawResult)    -> atom_to_list (RawResult);
		 true                -> RawResult
	     end,
    Result.

read_string (L) ->
    <<Length:32, R1/binary>> = L,
    <<Str:Length/binary, Rest/binary>> = R1,
    {binary_to_list (Str), Rest}.

read_strings (Buffer) ->
    {Str, R1} = read_string (Buffer),
    case Str of
	"" -> {[], R1};
	_  ->
	    {Strings, Rest} = read_strings (R1),
	    {[Str | Strings], Rest}
    end.

eval (Expr) ->
    {done, {ok, Tokens, _}, _} = erl_scan:tokens ("", Expr ++ ". ", 0),
    {ok, Exprs} = erl_parse:parse_exprs (Tokens),
    {value, Result, _} = erl_eval:exprs (Exprs, []),
    Result.

variables () ->
    case get (variables) of
	undefined ->
	    D = make_variables (get (env), dict:new ()),
	    put (variables, D),
	    D;
	D -> D
    end.    

find_variable (Name) ->
    dict:find (Name, variables ()).

make_variables ([], Dict) -> Dict;
make_variables ([NameValue | Others], Dict) ->
    {Name, Value} = split ("", NameValue),
    make_variables (Others, dict:store (Name, Value, Dict));
make_variables (undefined, Dict) -> Dict.

split (Name, [$= | Value]) -> {Name, Value};
split (Name, [Char | Value]) -> split (Name ++ [Char], Value).

inputs () ->
    case get (request) of
	undefined ->
	    D = parse_query (),
	    D;
	D -> D
    end.

find_input (Name) ->
    dict:find (Name, inputs ()).

parse_query () ->
    QS = case find_variable ("REQUEST_METHOD") of
	     {ok, "GET"} ->
		 case find_variable ("QUERY_STRING") of
		     {ok, V} -> V;
		     "" -> ""
		 end;
	     {ok, "POST"} ->
		 get (input);
	     {ok, "PUT"} ->
		 get (input);
	     _ ->
		 ""
	 end,
    dict:list_to_dict (httpd:parse_query (QS)).
		 
fetch_variable (Name) ->
    {ok, Value} = find_variable (Name),
    Value.

fetch_input (Name) ->
    {ok, Value} = find_input (Name),
    Value.

set_content_type (ContentType) ->
    put (contenttype, ContentType).

set_return_code (RC) ->
    put (return_code, RC).

set_return_code_immediate (RC) ->
    put (return_code, -RC).

lower ([]) -> [];
lower ([H | T]) when H >= $A, H =< $Z -> [H - $A + $a | lower (T)];
lower ([H | T]) -> [H | lower (T)].
    
get_cookies () ->
    case get (cookies) of
	undefined ->
	    Dict = case find_variable ("HTTP_COOKIE") of
		       {ok, Cookies} ->
			   dict:list_to_dict (httpd:parse_query (Cookies));
		       error ->
			   dict:new ()
		   end,
	    put (cookies, Dict),
	    Dict;
	X -> X
    end.

find_cookie (C) ->
    dict:find (C, get_cookies ()).

fetch_cookie (C) ->
    {ok, Answer} = find_cookie (C),
    Answer.

make_headers ([]) -> encapsulate ("");
make_headers ([{Name, Value} | T]) ->
    N = encapsulate (Name),
    V = encapsulate (Value),
    L = make_headers (T),
    << N/binary, V/binary, L/binary >>.

redirect (URL) ->
    set_return_code_immediate (302),
    add_header ("Location", URL).

add_header (Name, Value) ->
    put (headers, get (headers) ++ [{Name, Value}]).

set_only_headers () ->
    set_return_code_immediate (999).

self_url () ->
    "http://" ++
	case find_variable ("SERVER_NAME") of
	    {ok, Server} -> Server;
	    error        -> "dummy.example.com"
	end ++
	case find_variable ("SERVER_PORT") of
	    {ok, "80"} -> "";
	    {ok, Port} -> ":" ++ Port;
	    error      -> ""
	end ++
	case find_variable ("SCRIPT_NAME") of
	    {ok, Script} -> Script;
	    error        -> "/dummypath"
	end.

set_debug (D) ->
    eerl_handler_acceptor ! {set_debug, D}.
