-module (html).

-export ([text_html/0, flatten/1, dl/1, dd/1,
	  table/1, table/2, tr/1, tr/2, td/1, td/2, li/1, ul/1, href/1, href/2,
	  small/1, img/1, img/2, br/0, center/1, spaces/1, th/1, th/2,
	  h1/1, h2/1, h3/1, h4/1, h5/1, h6/1, font/2, hr/0,
	  ol/1, p/1, p/2, b/1, i/1, bi/1, title/1, html/2, divi/2,
	  head/1, body/1, body/2, stylesheet/1, form/3, meta_keywords/1,
	  meta_description/1, meta_refresh/2, button/3, select/2,
	  select/3, option/1, option/2, option/3, dict_parse_query/1]).

text_html () ->
    "Content-Type: text/html\n\n".

br () -> {"<br>", "", "", true, false}.
hr () -> {"<hr>", "", "", true, false}.

font (Text, Options) -> surround ("font", Text, Options).

center (Data) -> surround ("center", Data, [], true, false).

h1 (Heading) -> surround ("h1", Heading, [], true, false).
h2 (Heading) -> surround ("h2", Heading, [], true, false).
h3 (Heading) -> surround ("h3", Heading, [], true, false).
h4 (Heading) -> surround ("h4", Heading, [], true, false).
h5 (Heading) -> surround ("h5", Heading, [], true, false).
h6 (Heading) -> surround ("h6", Heading, [], true, false).
    
ol (List) -> surround ("ol", List, [], false, true).

p (Data) -> p (Data, []).
p (Data, Options) -> surround ("p", Data, Options, true, false).

divi (Data, Options) -> surround ("div", Data, Options, true, false).

b (Data) -> surround ("b", Data, []).
i (Data) -> surround ("i", Data, []).
bi (Data) -> b (i (Data)).

title (Title) -> surround ("title", Title, [], true, false).

html (Head, Body) -> surround ("html", [Head, Body], [], true, true).

head (Head) -> surround ("head", Head, [], true, true).

body (Body) -> body (Body, []).
body (Body, Options) -> surround ("body", Body, Options, true, true).

stylesheet (StyleSheet) ->
    format ("link", ["rel=\"stylesheet\"", "type=\"text/css\"",
		     "href=\"" ++ StyleSheet ++ "\""], true, true).

form (Method, Action, Body) ->
    surround ("form", Body, ["method=\"" ++ atom_to_list (Method) ++ "\"",
			     "action=\"" ++ Action ++ "\""]).

meta_name (Name, Content) ->
    format ("meta", ["name=\"" ++ Name ++ "\"",
		     "content=\"" ++ Content ++ "\""], true, true).

meta_equiv (Equiv, Content) ->
    format ("meta", ["http-equiv=\"" ++ Equiv ++ "\"",
		     "content=\"" ++ Content ++ "\""], true, true).

meta_keywords (Content) -> meta_name ("keywords", Content).

meta_description (Content) -> meta_name ("description", Content).

meta_refresh (Delay, URL) ->
   meta_equiv ("refresh", integer_to_list (Delay) ++ ";URL=" ++ URL).
    
button (Type, Name, Value) ->
    format ("input", ["type=\"" ++ atom_to_list (Type) ++ "\"",
		      "name=\"" ++ Name ++ "\"",
		      "value=\"" ++ Value ++ "\""]).

li (Item) -> surround ("li", Item, [], true, false).

ul (Items) -> surround ("ul", Items, [], false, true).

dl (Items) -> surround ("dl", Items, [], false, true).

dd (Item) -> surround ("dd", Item, [], true, false).

table (Content) ->
    table (Content, []).

table (Content, Options) ->
    surround ("table", Content, Options, false, true).

tr (Data) -> tr (Data, []).

tr (Data, Options) -> surround ("tr", Data, Options, true, true).

td (Data) -> td (Data, []).
td (Data, Options) -> surround ("td", Data, Options, true, false).

th (Data) -> th (Data, []).
th (Data, Options) -> surround ("th", Data, Options, true, false).

format (Tag, Options) ->
    lists:foldl (fun (X, Y) -> Y ++ " " ++ X end, "<" ++ Tag, Options) ++ ">".
    
format (Tag, Options, Outter, Inner) ->
    {format (Tag, Options), "", "", Outter, Inner}.

surround (Tag, Data, Options) ->
    surround (Tag, Data, Options, false, false).

surround (Tag, Data, Options, Outter, Inner) ->
    {format (Tag, Options), Data, "</" ++ Tag ++ ">", Outter, Inner}.

select (Name, Values) ->
    select (Name, Values, []).

select (Name, Values, Options) ->
    surround ("select", Values, ["name=\"" ++ Name ++ "\""] ++ Options,
	      false, true).

option (Text) ->
    option (Text, Text).

option (Text, Value) ->
    option (Text, Value, []).

option (Text, Value, Options) ->
    surround ("option", Text, ["value=\"" ++ Value ++ "\"" | Options], true,
	      false).

href ({Text, URL}) ->
    href ({Text, URL}, []).

href ({Text, URL}, Options) ->
    surround ("a", Text, ["href=\"" ++ URL ++ "\"" | Options]).

small (Data) -> surround ("small", Data, []).

spaces ([])            -> [];
spaces ([Head])        -> [Head];
spaces ([Head | Tail]) -> [Head, " " | spaces (Tail)].
			  
img (URL) -> img (URL, []).
img (URL, Options) ->
    format ("img", ["src=\"" ++ URL ++ "\"" | Options]).

flatten (Value) ->    
    flatten (Value, 0, true, true).

flatten ([], _, _, _) -> [];
flatten ({OTag, Data, CTag, Outter, Inner}, Indent, CanBreak, JustBroke) ->
    case (CanBreak or Outter) and (not JustBroke) of
	true  -> "\n" ++ xspaces (Indent);
	false -> ""
    end ++
	OTag ++
	flatten (Data, Indent + 1, Inner, false) ++
	case Inner of
	    true  -> "\n" ++ xspaces (Indent);
	    false -> ""
	end ++
	CTag;
flatten (List, Indent, CanBreak, JustBroke) when list (List) ->
    [flatten (L, Indent, CanBreak, false) || L <- List];
flatten (X, _, _, _) -> X.

xspaces (0) -> "";
xspaces (N) -> [$  | xspaces (N-1)].
    
dict_parse_query (Input) ->
    list_to_dict (httpd:parse_query (Input)).

list_to_dict ([]) -> dict:new ();
list_to_dict ([{K, V} | T]) ->
    Old = list_to_dict (T),
    case dict:is_key (K, T) of
	true  ->
	    dict:append (K, V, Old);
	false ->
	    dict:store (K, [V], Old)
    end.

    
