-module (http_security).
-export ([send_challenge/1, get_user/0, check_password/1,
	  require_authorization/3, require_mnesia_authorization/3,
	  require_simple_authorization/2,
	  require_mnesia_simple_authorization/2]).

%%
%% send_challenge (Realm) -> authorization_required
%%
%% Send a challenge with the appropriate HTTP return code with realm Realm
%%

send_challenge (Realm) ->
    BasicChallenge = "Basic realm=\"" ++ Realm ++ "\"",
    eerl_handler:add_header ("WWW-Authenticate", BasicChallenge),
    eerl_handler:set_return_code_immediate (401),
    authorization_required.

%%
%% get_user () -> User | undefined
%%
%% Get the user given in the HTTP Authorization header if any
%%

get_user () ->
    case get_user_password () of
	{basic, User, _Password} ->
	    User;
	undefined ->
	    undefined
    end.

%%
%% check_password (Password) -> ok | error
%%
%% Check that the given password matches Password
%%

check_password (Password) ->
    case get_user_password () of
	{basic, _User, Password} -> ok;
	_ -> error
    end.

%%
%% require_authorization (Realm, F, Data) -> {ok, User}
%%                                         | {ok, User, Opaque}
%%                                         | authorization_required
%%                                         | autorization_refused
%%
%% The callback function F/3 will be called with the name of the user and the
%% realm. Its profile is
%%
%% F (User, Realm, Data) -> ok
%%                       | {ok, Opaque}
%%                       | {password, Password}
%%                       | {password, Password, Opaque}
%%                       | retry
%%                       | refused
%%
%% ok and {ok, Opaque} validate the query. {password, Password} and
%% {password, Password, Opaque} mean that an extra verification must be done.
%%

require_authorization (Realm, F, Data) ->
    case get_user () of
	undefined ->
	    send_challenge (Realm);
	User ->
	    case F (User, Realm, Data) of
		ok ->
		    {ok, User};
		{ok, Opaque} ->
		    {ok, User, Opaque};
		{password, Password} ->
		    case check_password (Password) of
			ok    -> {ok, User};
			error -> send_challenge (Realm)
		    end;
		{password, Password, Opaque} ->
		    case check_password (Password) of
			ok    -> {ok, User, Opaque};
			error -> send_challenge (Realm)
		    end;
		retry ->
		    send_challenge (Realm);
		_ ->
		    refuse_authorization ()
	    end
    end.
		    
%%
%% require_mnesia_authorization (Realm, T, Data) -> {ok, User}
%%                                                | {ok, User, Opaque}
%%                                                | authorization_required
%%                                                | authorization_refused
%%
%% This is the same function, except that T will be called in the context of a mnesia
%% transaction.
%%

require_mnesia_authorization (Realm, T, Data) ->
    F = fun (User, _Realm, _Data) ->
		case mnesia:transaction (fun () -> T (User, Realm, Data) end) of
		    {atomic, Answer} -> Answer;
		    _                -> retry
		end
	end,
    require_authorization (Realm, F, Data).

%%
%% require_simple_authorization (Realm, [{User1, Password1}, ...]) -> {ok, User}
%%                                                                  | authorization_refused
%%
%% Check authorization against a simple user list
%%

require_simple_authorization (Realm, List) ->
    require_authorization (Realm, fun simple_authorization/3, List).

%%
%% require_mnesia_simple_authorization (Realm, Table) -> {ok, User}
%%                                                     | authorization_refused
%%
%% Check authorization in a mnesia table with two fields, user and password
%%

require_mnesia_simple_authorization (Realm, Table) ->
    require_mnesia_authorization (Realm, fun mnesia_check/3, Table).

%%
%% Internal functions
%%

simple_authorization (User, _Realm, List) ->
    case lists:keysearch (User, 1, List) of
	{value, {User, Password}} -> {password, Password};
	false                     -> retry
    end.

get_user_password () ->
    case eerl_handler:find_variable ("HTTP_AUTHORIZATION") of
	{ok, [$B, $a, $s, $i, $c, $ | UserPass64]} ->
	    UserPass = base64:decode (UserPass64),
	    {User, Pass} = split_first (UserPass),
	    {basic, User, Pass};
	_ ->
	    undefined
    end.

split_first ([]) -> {[], []};
split_first ([$: | T]) -> {[], T};
split_first ([H | T]) ->
    {A1, A2} = split_first (T),
    {[H | A1], A2}.

mnesia_check (User, _Realm, Table) ->
    case mnesia:match ({Table, User}) of
	[]                        -> retry;
	[{Table, User, Password}] -> {password, Password};
	_                         -> refused
    end.
	    
refuse_authorization () ->
    eerl_handler:set_return_code_immediate (403),
    authorization_refused.
