/*
 * mod_erl.c
 */

#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/mman.h>

#include "config.h"

#include "mod_erl.h"
#include "erl_scan.h"
#include "erl_marshal.h"

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "http_request.h"
#include "util_script.h"
#include "multithread.h"

#define MOD_ERL_STRING_VERSION (PACKAGE "/" VERSION)

typedef struct {
  int fd;
  size_t input_length;
  char *input;
  char **env;
} erl_conn;

typedef struct {
  char *host;
  unsigned int port;
} erl_config_rec;

typedef struct {
  void *start;
  int length;
} raw_data;

static int erl_handler (request_rec *);

static void *create_erl_dir_config (pool *, char *);
static void *merge_erl_dir_config (pool *, void *, void *);

static const char *erl_set_host (cmd_parms *, erl_config_rec *, const char *);
static const char *erl_set_port (cmd_parms *, erl_config_rec *, const char *);

static const handler_rec erl_handlers[] = {
  {"application/x-httpd-erl", erl_handler}
};

static void erl_startup (server_rec *, pool *);
static int erl_call (char *begin, char *end, erl_option *options,
		     request_rec *r, erl_conn *ec, array_header *returns);

static const command_rec erl_cmds[] =
{
  {"ErlHost", erl_set_host, NULL,
   OR_FILEINFO, RAW_ARGS,
   "the host to use when evaluating an Erlang command."},
  {"ErlPort", erl_set_port, NULL,
   OR_FILEINFO, RAW_ARGS,
   "the port to use when evaluating an Erlang command."},
  {NULL}
};

MODULE_VAR_EXPORT module erl_module = {
    STANDARD_MODULE_STUFF,
    erl_startup,                /* initializer */
    create_erl_dir_config,      /* dir config creater */
    merge_erl_dir_config,       /* dir merger --- default is to override */
    NULL,                       /* create per-server config structure */
    NULL,                       /* merge server config */
    erl_cmds,                   /* command table */
    erl_handlers,               /* handlers */
    NULL,                       /* filename translation */
    NULL,                       /* check_user_id */
    NULL,                       /* check auth */
    NULL,                       /* check access */
    NULL,                       /* type_checker */
    NULL,                       /* fixups */
    NULL,                       /* logger */
    NULL,                       /* header parser */
    NULL,                       /* child_init */
    NULL,                       /* child_exit */
    NULL,                       /* post read-request */
#ifdef EAPI
    NULL,                       /* add_module */
    NULL,                       /* remove_module */
    NULL,                       /* rewrite_command */
    NULL,                       /* new_connection */
    NULL                        /* close_connection */
#endif /* EAPI */
};

static void
erl_startup (server_rec *s, pool *p)
{
  erl_server_config *conf =
    (erl_server_config *) ap_get_module_config (s->module_config, &erl_module);
#if MODULE_MAGIC_NUMBER >= 19980507
  ap_add_version_component (MOD_ERL_STRING_VERSION);
#endif
}

static void *
create_erl_dir_config(pool *p, char *d)
{
  erl_config_rec *e =
    (erl_config_rec *) ap_palloc (p, sizeof (erl_config_rec));
  e->host = NULL;
  e->port = 0;
  return e;
}

static void *
merge_erl_dir_config(pool *p, void *ubase, void *uadd)
{
  erl_config_rec *base = (erl_config_rec *) ubase;
  erl_config_rec *add = (erl_config_rec *) uadd;
  erl_config_rec *new =
    (erl_config_rec *) ap_palloc (p, sizeof (erl_config_rec));
  new->host = add->host ? add->host : base->host;
  new->port = add->port ? add->port : base->port;
  return new;
}

static const char *
erl_set_host (cmd_parms *cmd, erl_config_rec *e, const char *arg)
{
  size_t len = strlen (arg) + 1;
  e->host = ap_palloc (cmd->pool, len);
  memcpy (e->host, arg, len);
  return NULL;
}

static const char *
erl_set_port (cmd_parms *cmd, erl_config_rec *e, const char *arg)
{
  e->port = atoi (arg);
  return NULL;
}

static int
erl_handler (request_rec *r)
{
  int fd;
  char *file, *index;
  int rc, i;
  int return_code = OK;
  array_header *returns;
  raw_data *rd;
  erl_conn ec;
  int skip_body = 0;

  if (r->finfo.st_mode == 0) {
    return NOT_FOUND;
  }

  if (S_ISDIR (r->finfo.st_mode))
    return FORBIDDEN;

  r->content_type = NULL;
  returns = ap_make_array (r->pool, 3, sizeof (raw_data));

  ap_soft_timeout ("mod_erl parsing", r);

  fd = open (r->filename, O_RDONLY);
  if (fd < 0) {
    return FORBIDDEN;
  }

  file = mmap (NULL, r->finfo.st_size + 1, PROT_READ | PROT_WRITE, 
	       MAP_PRIVATE, fd, 0);
  if (file == MAP_FAILED) {
    close (fd);
    return FORBIDDEN;
  }
  file[r->finfo.st_size] = '\0';

  ec.fd = -1;

  for (index = file; index - file <= r->finfo.st_size;) {
    char *begin, *end, *content_begin, *content_end;
    erl_option *options;

    if (lookup_erl (index, &begin, &end, &content_begin, &content_end,
		    &options)) {
      if (begin > index && !r->header_only) {
	rd = (raw_data *) ap_push_array (returns);
	rd->start = index;
	rd->length = begin - index;
      }

      /* Call Erlang */
      ap_kill_timeout (r);
      rc = erl_call (content_begin, content_end, options,  r, &ec, returns);
      if (rc < 0) {
	if (rc != -1) {
	  if (rc != -999) {
	    return_code = -rc;
	  }
	  skip_body = 1;
	  break;
	}
	munmap (file, r->finfo.st_size);
	close (fd);
	return FORBIDDEN;
      }

      if (rc) {
	return_code = rc;
      }
      
      ap_soft_timeout ("parsing file", r);
      index = end + 1;
    } else {
       if (!r->header_only) {
	 /* End of file */
	 rd = (raw_data *) ap_push_array (returns);
	 rd->start = index;
	 rd->length = file + r->finfo.st_size - index;
       }
      break;
    }
  }

  if (!r->content_type && return_code == OK) {
    r->content_type = "text/html";
  }

  if (return_code != OK) {
    r->status = return_code;
  }
  ap_send_http_header (r);
  if (!r->header_only && !skip_body) {
    for (i = 0; i < returns->nelts; i++) {
      rd = (raw_data *) returns->elts;
      ap_rwrite (rd[i].start, rd[i].length, r);
    }
  }
  ap_kill_timeout (r);
  munmap (file, r->finfo.st_size);
  close (fd);
  if (ec.fd != -1) {
    shutdown (ec.fd, SHUT_RDWR);
  }
  return OK;
}

int
erl_call (char *begin, char *end, erl_option *options,
	  request_rec *r, erl_conn *ec, array_header *returns)
{
  erl_message msg;
  char **env;
  erl_config_rec *e = ap_get_module_config (r->per_dir_config, &erl_module);
  const char *result;
  size_t length;
  int i, rc, retval;
  raw_data *rd;

  if (ec->fd == -1) {
    /* No connexion has been opened at this time */
    struct hostent *h;
    struct sockaddr_in sin;
    char *host;
    unsigned int port;

    host = e->host ? e->host : MOD_ERL_DEFAULT_HOST;
    port = e->port ? e->port : MOD_ERL_DEFAULT_PORT;

    h = ap_pgethostbyname (r->pool, host);
#ifdef HAVE_SIN_LEN
    sin.sin_len = 4;
#endif
    sin.sin_family = AF_INET;
    memcpy (&sin.sin_addr, h->h_addr_list[0], 4);
    sin.sin_port = htons (port);
    
    ec->fd = ap_psocket (r->pool, PF_INET, SOCK_STREAM, 0);
    if (ec->fd < 0) {
      ap_log_rerror (APLOG_MARK, APLOG_ERR, r,
		     "cannot create socket");
      return -1;
    }
    if (connect (ec->fd, (struct sockaddr *)&sin, sizeof sin) < 0) {
      ap_log_rerror (APLOG_MARK, APLOG_ERR, r,
		     "cannot connect to Erlang node");
      return -1;
    }

    /* If stdin must be read, read it now */
    if ((r->method_number == M_POST) || (r->method_number == M_PUT)) {
      const char *ascii_length =
	ap_table_get (r->headers_in, "Content-length");

      if (!ascii_length) {
	ap_log_rerror (APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
		       "No content-length found for %s method", r->method);
	return -1;
      }

      ec->input_length = atoi (ascii_length);
      ec->input = ap_pcalloc (r->pool, ec->input_length + 1);

      /* Transfer any put/post args, CERN style...
       * Note that we already ignore SIGPIPE in the core server.
       * Copied from mod_cgi.c
       */

      if ((retval = ap_setup_client_block (r, REQUEST_CHUNKED_ERROR))) {
	return retval;
      }

      if (ap_should_client_block (r)) {
        int dbsize, len_read;
	size_t dbpos = 0;
	
        ap_hard_timeout("copy script args", r);
        while
	  ((dbpos < ec->input_length) &&
	   ((len_read = ap_get_client_block (r, &ec->input[dbpos],
					     ec->input_length - dbpos)) > 0)) {
	  dbpos += len_read;
	}
	ap_kill_timeout(r);

	if (dbpos != ec->input_length) {
	  ap_log_rerror (APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
			 "%d bytes read from client, %d expected", dbpos,
			 ec->input_length);
	  return -1;
	}
      }
    } else {
      ec->input = "";
      ec->input_length = 0;
    }

    ap_add_common_vars (r);
    ap_add_cgi_vars (r);
    ec->env = ap_create_environment (r->pool, r->subprocess_env);
  }

  init_message (&msg);
  
  /* First send the standard input */
  add_raw (&msg, ec->input, ec->input_length, r);

  /* Send the environment then an empty string */
  for (i = 0; ec->env[i]; i++) {
    add_string (&msg, ec->env[i], r);
  }
  add_string (&msg, "", r);

  /* Send the code itself */
  add_raw (&msg, begin, end - begin + 1, r);

  /* Send the message and wait for the answer */
  ap_soft_timeout ("call erlang", r);
  if (send_message (ec->fd, &msg, 1, r)) {
    ap_kill_timeout (r);
    ap_log_rerror (APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
		   "error while sending data to Erlang node");
    return -1;
  }

  /* Read a message (with its length) from the socket */
  if (read_message (ec->fd, &msg, 0, r)) {
    ap_kill_timeout (r);
    ap_log_rerror (APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
		   "error while receiving data from Erlang node");
    return -1;
  }

  /* Get the return code, if any */
  rc = read_integer (&msg);

  /* Get the string representing the content-type and copy it if there is one */
  result = read_raw (&msg, &length);
  if (length) {
    r->content_type = ap_pcalloc (r->pool, length + 1);
    memcpy ((char *) r->content_type, result, length);
  }

  /* Read strings representing extra headers to add to the result */
  for (;;) {
    result = read_string (&msg, r);
    if (!*result) {
      break;
    }
    ap_table_setn (r->headers_out, result, read_string (&msg, r));
  }

  /* Read the result and add it if we are not in headers only mode */
  result = read_raw (&msg, &length);
  if (!r->header_only) {
    rd = (raw_data *) ap_push_array (returns);
    rd->start = ap_palloc (r->pool, length);
    memcpy (rd->start, result, length);
    rd->length = length;
  }

  ap_kill_timeout (r);

  return rc;
}
